package org.artifactory.metrics.providers.features;

import com.google.gson.Gson;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.addon.retention.ArtifactRetentionStats;
import org.artifactory.api.callhome.FeatureGroup;
import org.artifactory.util.UnsupportedByLicenseException;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoSession;
import org.mockito.quality.Strictness;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import static org.mockito.Mockito.when;

/**
 * @author Mor Merhav
 */
public class ArtifactLifecycleFeatureTest {

    private MockitoSession mockitoSession;
    private ArtifactRetentionFeature artifactRetentionFeature;

    @Mock
    private AddonsManager addonsManager;
    @Mock
    private ArtifactRetentionAddon artifactRetentionAddon;

    @BeforeMethod
    public void setUp() {
        mockitoSession = Mockito.mockitoSession()
                .strictness(Strictness.STRICT_STUBS)
                .initMocks(this)
                .startMocking();

        artifactRetentionFeature = new ArtifactRetentionFeature(addonsManager);
        when(addonsManager.addonByType(ArtifactRetentionAddon.class)).thenReturn(artifactRetentionAddon);
    }

    @AfterMethod
    public void tearDown() {
        mockitoSession.finishMocking();
    }

    @Test
    public void retentionDisabledNFeatureReturnsNull() {
        when(artifactRetentionAddon.isRetentionEnabled()).thenReturn(false);
        Assert.assertNull(artifactRetentionFeature.getFeature());
    }

    @Test
    public void isColdInstanceTrueFeatureReturnsColdInstanceField() {
        when(artifactRetentionAddon.isRetentionEnabled()).thenReturn(true);
        when(artifactRetentionAddon.isColdInstance()).thenReturn(true);

        FeatureGroup feature = artifactRetentionFeature.getFeature();
        String featureJson = new Gson().toJson(feature);
        Assert.assertEquals(featureJson,
                "{\"name\":\"Artifactory On Ice\",\"attributes\":{\"Is_Cold_Instance\":true}}");
    }

    @Test
    public void artifactLifecycleStatsNullFeatureReturnsNull() {
        when(artifactRetentionAddon.isRetentionEnabled()).thenReturn(true);
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        when(artifactRetentionAddon.getArtifactRetentionStats()).thenReturn(null);
        Assert.assertNull(artifactRetentionFeature.getFeature());
    }

    @Test
    public void artifactLifecycleFeatureReturnsAllStats() {
        when(artifactRetentionAddon.isRetentionEnabled()).thenReturn(true);
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);

        ArtifactRetentionStats artifactRetentionStats = new ArtifactRetentionStats(
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11);
        when(artifactRetentionAddon.getArtifactRetentionStats()).thenReturn(artifactRetentionStats);
        FeatureGroup feature = artifactRetentionFeature.getFeature();
        String featureJson = new Gson().toJson(feature);
        Assert.assertEquals(featureJson,
                "{\"name\":\"Artifactory On Ice\",\"attributes\":{\"Is_Cold_Instance\":false,\"Total_Policies\":1," +
                        "\"Last_Day_Archive_Policies_Runs\":2,\"Last_Day_Archived_Artifacts\":3," +
                        "\"Last_Day_Archived_Artifacts_Size\":5,\"Last_Day_Archived_Artifacts_Failed\":4," +
                        "\"Last_Day_Archived_Artifacts_Failed_Size\":6,\"Last_Day_Restore_Runs\":7," +
                        "\"Last_Day_Restored_Artifacts\":8,\"Last_Day_Restored_Artifacts_Size\":10," +
                        "\"Last_Day_Restored_Artifacts_Failed\":9,\"Last_Day_Restored_Artifacts_Failed_Size\":11}}");
    }

    @Test
    public void getStatsThrowsExceptionArtifactLifecycleFeatureReturnsNull() {
        when(artifactRetentionAddon.isRetentionEnabled()).thenReturn(true);
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        when(artifactRetentionAddon.getArtifactRetentionStats()).thenThrow(new UnsupportedByLicenseException("msg"));

        Assert.assertNull(artifactRetentionFeature.getFeature());
    }


}