package org.artifactory.schedule.monitor;

import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.message.BasicStatusLine;
import org.artifactory.common.ConstantValues;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jetbrains.annotations.NotNull;
import org.testng.Assert;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.atLeast;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Alexei Vainshtein
 */
public class OnlineMonitorServiceImplTest extends ArtifactoryHomeBoundTest {

    @Test
    public void testAddOfflineRepo() {
        OnlineMonitorServiceImpl onlineMonitorService = spy(new OnlineMonitorServiceImpl());
        RemoteRepo repo = prepareRemoteRepo();
        addOfflineRepo(onlineMonitorService, repo);
    }

    @Test
    public void testAddOfflineRepoTwice() {
        OnlineMonitorServiceImpl onlineMonitorService = spy(new OnlineMonitorServiceImpl());
        RemoteRepo repo = prepareRemoteRepo();
        ScheduledExecutorService scheduledExecutorService = mock(ScheduledExecutorService.class);
        ScheduledFuture oldFuture = mock(ScheduledFuture.class);
        doAnswer(invocation -> {
            OnlineMonitorServiceImpl.OnlineMonitorRunnable argument =
                    invocation.getArgument(0, OnlineMonitorServiceImpl.OnlineMonitorRunnable.class);
            assertEquals(argument.calculateNextOnlineCheckDelay(), 100);
            return oldFuture;
        }).when(scheduledExecutorService)
                .schedule(any(Runnable.class), eq(0L), eq(TimeUnit.MILLISECONDS));

        onlineMonitorService.onlineMonitorScheduler = scheduledExecutorService;
        onlineMonitorService.createOfflineRepoMonitor(repo);
        onlineMonitorService.createOfflineRepoMonitor(repo);
        verify(scheduledExecutorService, times(1))
                .schedule(any(Runnable.class), eq(0L), eq(TimeUnit.MILLISECONDS));
        assertTrue(onlineMonitorService.offlineReposFutures.containsKey(repo.getKey()));
        assertEquals(onlineMonitorService.offlineReposFutures.get(repo.getKey()), oldFuture);
    }

    @Test(enabled = false, description = "Fails on pipelines, pass locally.")
    public void testAddOfflineRepoTwiceSimultaneously() {
        CountDownLatch latch = new CountDownLatch(1);
        int threads = 70;
        ExecutorService service = Executors.newFixedThreadPool(threads);
        OnlineMonitorServiceImpl onlineMonitorService = spy(new OnlineMonitorServiceImpl());
        ScheduledFuture scheduledFuture = mock(ScheduledFuture.class);
        RemoteRepo repo = prepareRemoteRepo();
        ScheduledExecutorService scheduledExecutorService = prepareExecutor(onlineMonitorService, scheduledFuture);
        List<Future<Boolean>> futures = new ArrayList<>();
        for (int i = 0; i < threads; i++) {
            futures.add(service.submit(() -> {
                latch.await();
                onlineMonitorService.createOfflineRepoMonitor(repo);
                return true;
            }));
        }
        latch.countDown();
        try {
            futures.forEach(future -> {
                try {
                    future.get();
                } catch (InterruptedException | ExecutionException e) {
                    Assert.fail("Future failed with: " + e.getMessage(), e);
                }
            });
        } finally {
            service.shutdown();
        }
        verify(scheduledExecutorService, atLeast(2))
                .schedule(any(Runnable.class), eq(0L), eq(TimeUnit.MILLISECONDS));
        assertTrue(onlineMonitorService.offlineReposFutures.containsKey(repo.getKey()));
        assertEquals(onlineMonitorService.offlineReposFutures.get(repo.getKey()), scheduledFuture);
        verify(scheduledFuture, atLeast(1)).cancel(true);
    }

    @Test
    public void testOnlineMonitorScheduler() {
        OnlineMonitorServiceImpl onlineMonitorService = new OnlineMonitorServiceImpl();
        try {
            onlineMonitorService.initOnlineMonitorScheduler();
            assertEquals(
                    ((ScheduledThreadPoolExecutor) onlineMonitorService.onlineMonitorScheduler).getMaximumPoolSize(),
                    ConstantValues.onlineMonitorThreadsCount.getInt());
            assertTrue(((ScheduledThreadPoolExecutor) onlineMonitorService.onlineMonitorScheduler)
                    .allowsCoreThreadTimeOut());
            assertTrue(((ScheduledThreadPoolExecutor) onlineMonitorService.onlineMonitorScheduler)
                    .getRemoveOnCancelPolicy());
            assertEquals(((ScheduledThreadPoolExecutor) onlineMonitorService.onlineMonitorScheduler)
                    .getKeepAliveTime(TimeUnit.SECONDS), 60);
        } finally {
            onlineMonitorService.destroy();
            assert onlineMonitorService.onlineMonitorScheduler != null;
            assertTrue(onlineMonitorService.onlineMonitorScheduler.isShutdown());
        }
    }

    @Test(dataProvider = "statusProvider")
    public void testIsResourceUnavailable(int statusCode, boolean isUnavailable) {
        OnlineMonitorServiceImpl onlineMonitorService = new OnlineMonitorServiceImpl();
        BasicStatusLine basicStatusLine = mock(BasicStatusLine.class);
        when(basicStatusLine.getStatusCode()).thenReturn(statusCode);
        assertEquals(onlineMonitorService.isResourceUnavailable(basicStatusLine), isUnavailable);
    }

    @Test
    public void testCalculateNextOnlineCheckDelay() {
        RemoteRepo repo = prepareRemoteRepo();
        OnlineMonitorServiceImpl onlineMonitorService = new OnlineMonitorServiceImpl();
        OnlineMonitorServiceImpl.OnlineMonitorRunnable runnable = onlineMonitorService.createRunnable(repo);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 100);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 100);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 100);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 155);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 547);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 1933);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 6823);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 24082);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 84998);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 300000);
        assertEquals(runnable.calculateNextOnlineCheckDelay(), 300000);
    }

    @Test(dataProvider = "onlineProvider")
    public void testCheckOnline(int statusCode, boolean isOffline) throws IOException {
        RemoteRepo repo = prepareRemoteRepo();
        OnlineMonitorServiceImpl onlineMonitorService = new OnlineMonitorServiceImpl();
        OnlineMonitorServiceImpl.OnlineMonitorRunnable runnable = onlineMonitorService.createRunnable(repo);
        prepareCheckOnline(statusCode, repo);
        assertEquals(runnable.checkOnline(), !isOffline);
    }

    @Test
    public void testCheckOnlineWithException() throws IOException {
        RemoteRepo repo = prepareRemoteRepo();
        OnlineMonitorServiceImpl onlineMonitorService = new OnlineMonitorServiceImpl();
        OnlineMonitorServiceImpl.OnlineMonitorRunnable runnable = onlineMonitorService.createRunnable(repo);
        when(repo.getUrl()).thenReturn("https://this.url");
        CloseableHttpClient closeableHttpClient = mock(CloseableHttpClient.class);
        when(closeableHttpClient.execute(any(HttpGet.class))).thenThrow(new IOException("nope"));
        when(repo.getHttpClient()).thenReturn(closeableHttpClient);
        assertFalse(runnable.checkOnline());
    }

    @Test
    public void testStopMonitorForRepo() {
        OnlineMonitorServiceImpl onlineMonitorService = spy(new OnlineMonitorServiceImpl());
        RemoteRepo repo = prepareRemoteRepo();
        addOfflineRepo(onlineMonitorService, repo);
        ScheduledFuture scheduledFuture = onlineMonitorService.offlineReposFutures.get("key");
        onlineMonitorService.stopMonitorForRepo(repo);
        verify(scheduledFuture, times(1)).cancel(true);
        verify(onlineMonitorService, times(1)).removeRepoFromOfflineMap(repo.getKey());
    }

    @Test
    public void testRunSuccess() throws IOException {
        RemoteRepo repo = prepareRemoteRepo();
        when(repo.isAssumedOffline()).thenReturn(true);
        OnlineMonitorServiceImpl onlineMonitorService = spy(new OnlineMonitorServiceImpl());
        OnlineMonitorServiceImpl.OnlineMonitorRunnable runnable = onlineMonitorService.createRunnable(repo);
        prepareCheckOnline(202, repo);
        runnable.run();
        verify(repo, times(1)).resetAssumedOffline();
        verify(onlineMonitorService, times(1)).removeRepoFromOfflineMap(repo.getKey());
    }

    @Test
    public void testRunFailed() throws IOException {
        OnlineMonitorServiceImpl onlineMonitorService = spy(new OnlineMonitorServiceImpl());
        RemoteRepo repo = prepareRemoteRepo();
        ScheduledFuture newFuture = mock(ScheduledFuture.class);
        ScheduledFuture oldFuture = mock(ScheduledFuture.class);
        onlineMonitorService.offlineReposFutures.put(repo.getKey(), oldFuture);
        prepareExecutor(100L, onlineMonitorService, newFuture);
        when(repo.isAssumedOffline()).thenReturn(false);
        OnlineMonitorServiceImpl.OnlineMonitorRunnable runnable = onlineMonitorService.createRunnable(repo);
        prepareCheckOnline(502, repo);
        runnable.run();
        verify(repo, times(1)).setAssumedOffline();
        verify(onlineMonitorService, times(0)).removeRepoFromOfflineMap(repo.getKey());
    }

    private void addOfflineRepo(OnlineMonitorServiceImpl onlineMonitorService, RemoteRepo repo) {
        ScheduledFuture scheduledFuture = mock(ScheduledFuture.class);
        ScheduledExecutorService scheduledExecutorService =
                prepareExecutor(onlineMonitorService, scheduledFuture);
        onlineMonitorService.createOfflineRepoMonitor(repo);
        verify(scheduledExecutorService, times(1))
                .schedule(any(Runnable.class), eq(0L), eq(TimeUnit.MILLISECONDS));
        assertTrue(onlineMonitorService.offlineReposFutures.containsKey(repo.getKey()));
        assertEquals(onlineMonitorService.offlineReposFutures.get(repo.getKey()), scheduledFuture);
    }

    private ScheduledExecutorService prepareExecutor(OnlineMonitorServiceImpl onlineMonitorService,
            ScheduledFuture scheduledFuture) {
        return prepareExecutor(0, onlineMonitorService, scheduledFuture);
    }

    private ScheduledExecutorService prepareExecutor(long initDelay, OnlineMonitorServiceImpl onlineMonitorService,
            ScheduledFuture scheduledFuture) {
        ScheduledExecutorService scheduledExecutorService = mock(ScheduledExecutorService.class);
        doAnswer(invocation -> {
            OnlineMonitorServiceImpl.OnlineMonitorRunnable argument =
                    invocation.getArgument(0, OnlineMonitorServiceImpl.OnlineMonitorRunnable.class);
            assertEquals(argument.calculateNextOnlineCheckDelay(), 100);
            return scheduledFuture;
        }).when(scheduledExecutorService)
                .schedule(any(Runnable.class), eq(initDelay), eq(TimeUnit.MILLISECONDS));

        onlineMonitorService.onlineMonitorScheduler = scheduledExecutorService;
        return scheduledExecutorService;
    }

    @NotNull
    private RemoteRepo prepareRemoteRepo() {
        RemoteRepo repo = mock(RemoteRepo.class);
        when(repo.getKey()).thenReturn("key");
        when(repo.getAssumedOfflinePeriodSecs()).thenReturn(300L);
        return repo;
    }

    private void prepareCheckOnline(int statusCode, RemoteRepo repo) throws IOException {
        when(repo.getUrl()).thenReturn("https://this.url");
        CloseableHttpClient closeableHttpClient = mock(CloseableHttpClient.class);
        CloseableHttpResponse response = mock(CloseableHttpResponse.class);
        BasicStatusLine statusLine = mock(BasicStatusLine.class);
        when(statusLine.getReasonPhrase()).thenReturn("Reason");
        when(statusLine.getStatusCode()).thenReturn(statusCode);
        when(response.getStatusLine()).thenReturn(statusLine);
        when(closeableHttpClient.execute(any(HttpGet.class))).thenReturn(response);
        when(repo.getHttpClient()).thenReturn(closeableHttpClient);
    }

    @DataProvider
    public Object[][] statusProvider() {
        return new Object[][] {
                {503, true},
                {502, true},
                {504, true},
                {505, true},
                {400, false},
                {500, false},
                {200, false},
                {202, false},
                {201, false},
                {301, false},
                {306, false},
        };
    }

    @DataProvider
    public Object[][] onlineProvider() {
        return new Object[][] {
                {503, true},
                {502, true},
                {504, true},
                {505, true},
                {400, false},
                {500, false},
                {200, false},
                {202, false},
                {201, false},
                {301, false},
                {306, false},
        };
    }
}