package org.artifactory.rest.resource.servicetrust;

import com.google.common.annotations.VisibleForTesting;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.authentication.AuthenticationToken;
import org.artifactory.rest.common.util.RestUtils;
import org.artifactory.servicetrust.ExchangeTokenRequest;
import org.artifactory.servicetrust.ExchangeTokenResponse;
import org.artifactory.servicetrust.ServiceTrustException;
import org.artifactory.servicetrust.ServiceTrustService;
import org.artifactory.servicetrust.ServiceTrustUseCase;
import org.artifactory.util.HttpUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.security.RolesAllowed;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * @author igoru
 */
@Component
@Path(ServiceTrustResource.BASE_PATH)
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class ServiceTrustResource {
    private static final Logger log = LoggerFactory.getLogger(ServiceTrustResource.class);
    private static final String PAIRING = "pairing";
    private static final String PAIRINGS = "pairings";
    private static final String EXCHANGE = "exchange";
    private static final String ERROR_CREATING_PAIRING_TOKEN_PATTERN = "Error creating pairing token. %s";
    private static final String ERROR_GETTING_PAIRING_TOKEN_LIST_PATTERN = "Error getting list of pairing tokens. %s";
    private static final String ERROR_TOKEN_PAIRING_PATTERN = "Error while token pairing. %s";
    private static final String ERROR_TOKEN_EXCHANGE_PATTERN = "Error while token exchange. %s";
    private static final String UNKNOWN_USE_CASE = "Unknown use case";

    public static final String TRUST_WITH_SERVICE_WAS_ESTABLISHED_SUCCESSFULLY =
            "Trust with service %s was established successfully";

    public static final String BASE_PATH = "/v1/service_trust/";

    @Autowired
    private ServiceTrustService serviceTrustService;

    @Autowired
    private CentralConfigService configService;

    @Context
    private HttpServletRequest request;

    @GET
    @RolesAllowed({AuthorizationService.ROLE_ADMIN})
    @Path(PAIRINGS + "/{usecase: .*}")
    @Produces(MediaType.APPLICATION_JSON)
    public Response listTokens(@PathParam("usecase") String useCasePath) {
        Optional<ServiceTrustUseCase> useCase = serviceTrustService.findUseCase(useCasePath);
        return useCase.map(u -> createTokenListResponse(useCasePath))
                .orElse(Response.status(Response.Status.NOT_FOUND)
                        .entity(new MessageResponse(UNKNOWN_USE_CASE)).build());
    }

    @POST
    @RolesAllowed({AuthorizationService.ROLE_ADMIN})
    @Path(PAIRING + "/{usecase: .*}")
    @Produces(MediaType.APPLICATION_JSON)
    public Response createToken(@PathParam("usecase") String useCasePath) {
        Optional<ServiceTrustUseCase> useCase = serviceTrustService.findUseCase(useCasePath);
        return useCase.map(u -> createTokenResponse(useCasePath, u))
                .orElse(Response.status(Response.Status.NOT_FOUND)
                        .entity(new MessageResponse(UNKNOWN_USE_CASE)).build());
    }

    @PUT
    @RolesAllowed({AuthorizationService.ROLE_ADMIN})
    @Path(PAIRING + "/{usecase: .*}")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public Response registerToken(@PathParam("usecase") String useCasePath, PairingResponse body) {
        Optional<ServiceTrustUseCase> useCase = serviceTrustService.findUseCase(useCasePath);
        return useCase.map(u -> pairingTokenResponse(useCasePath, u, body.getPairingToken()))
                .orElse(Response.status(Response.Status.NOT_FOUND)
                        .entity(new MessageResponse(UNKNOWN_USE_CASE)).build());
    }

    @POST
    @Path(EXCHANGE)
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public Response exchangeToken(ExchangeTokenRequest exchangeTokenRequest) {
        try {
            ExchangeTokenResponse response = serviceTrustService.exchangeToken(exchangeTokenRequest);
            return Response.ok(response).build();
        } catch (Exception e) {
            return handleException(String.format(ERROR_TOKEN_EXCHANGE_PATTERN, e.getMessage()), e);
        }
    }

    @VisibleForTesting
    void setServiceTrustService(ServiceTrustService serviceTrustService) {
        this.serviceTrustService = serviceTrustService;
    }

    @VisibleForTesting
    String getRestApiUrl() {
        return RestUtils.getRestApiUrl(request) + BASE_PATH;
    }

    @VisibleForTesting
    String getBaseUrl() {
        return HttpUtils.getPlatformServletContextUrl(request);
    }

    private Response createTokenResponse(String useCasePath, ServiceTrustUseCase serviceTrustUseCase) {
        try {
            String baseUrl = getBaseUrl();
            if (useCasePath.contains("federated-repo/")) {
                String fedBaseUrl = configService.getDescriptor().getFederatedRepoUrlBaseOrDefault();
                if(StringUtils.isNotBlank(fedBaseUrl)) {
                    baseUrl = fedBaseUrl.replace("/artifactory","");
                }
            }

            String restApiUrl = getRestApiUrl();
            String registerUrl = restApiUrl + PAIRING;
            String exchangeUrl = restApiUrl + EXCHANGE;
            String token = serviceTrustService
                    .getPairingToken(useCasePath, serviceTrustUseCase.getTrust(), baseUrl, registerUrl, exchangeUrl);
            PairingResponse response = new PairingResponse(token);
            return Response.ok().entity(response).build();
        } catch (Exception e) {
            return handleException(String.format(ERROR_CREATING_PAIRING_TOKEN_PATTERN, e.getMessage()), e);
        }
    }

    private Response createTokenListResponse(String useCasePath) {
        try {
            String baseUrl = getBaseUrl();
            Collection<AuthenticationToken> tokens = serviceTrustService.getPairingTokenList(useCasePath, baseUrl);

            List<BaseUrlModel> pairedInstances = tokens.stream()
                    .map(AuthenticationToken::getBaseUrl)
                    .map(BaseUrlModel::new)
                    .collect(Collectors.toList());

            return Response.ok().entity(pairedInstances).build();

        } catch (Exception e) {
            return handleException(String.format(ERROR_GETTING_PAIRING_TOKEN_LIST_PATTERN, e.getMessage()), e);
        }
    }


    private Response pairingTokenResponse(String useCasePath, ServiceTrustUseCase serviceTrustUseCase, String token) {
        try {
            serviceTrustService.pairToken(serviceTrustUseCase, token);
            String message =
                    String.format(TRUST_WITH_SERVICE_WAS_ESTABLISHED_SUCCESSFULLY, serviceTrustUseCase.getName());
            return Response.ok(new MessageResponse(message)).build();
        } catch (Exception e) {
            return handleException(String.format(ERROR_TOKEN_PAIRING_PATTERN, e.getMessage()), e);
        }
    }

    private Response handleException(String message, Exception e) {
        log.error(message, e);
        MessageResponse messageResponse = new MessageResponse(message);
        if (e instanceof ServiceTrustException || e instanceof IllegalStateException) {
            return Response.status(Response.Status.BAD_REQUEST).entity(messageResponse).build();
        } else {
            return Response.serverError().entity(messageResponse).build();
        }
    }

}
